-- ************************************************************************** --
-- * TitanWG.lua                                                            * --
-- *                                                                        * --
-- * By: Rothpada                                                           * --
-- * Translations by: Gomntep (deDE), LigthSora (esES,esMX),                * -- 
-- *                  StingerSoft (ruRU)                                    * --
-- ************************************************************************** --

local TITAN_WG_ID = "TitanWG";
local AceTimer = LibStub("AceTimer-3.0");
local L = LibStub("AceLocale-3.0"):GetLocale("TitanWG", true)

local TitanWG_NORTHREND_INDEX = 4;
local TitanWG_WINTERGRASP_INDEX = 10; 
local TitanWG_MAX_BATTLE_LENGTH = 30*60;
local TitanWG_MAX_TIME_BETWEEN_BATTLES = 2.5*3600;
local TitanWG_WARNING_TIME = 10*60;
local TitanWG_LEVEL_THRESHOLD = 80;

local TitanWG_IS_DEBUGGING = ( false and GetRealmName() == "Shadow Council" ) or UnitName("player") == "Balther";

-- TODO: Turn into some sort of enumeration
local TitanWG_STATE_CURRENTLY_IN_NORTHREND = 0;
local TitanWG_STATE_RECENTLY_IN_NORTHREND = 1;
local TitanWG_STATE_BATTLE_UNDERWAY = 2;
local TitanWG_STATE_NOT_RECENTLY_IN_NORTHREND = 3;
local TitanWG_STATE_NOT_BEEN_TO_NORTHREND = 4;

local TitanWG_announced = false;
local TitanWG_continent = -1;
local TitanWG_owner = nil;
local TitanWG_player = nil;
local TitanWG_realm = nil;

-- The state machine's struct
local TitanWG_State = {};
	TitanWG_State.isInfoAccurate = function() return TitanWG_State.state <= TitanWG_STATE_RECENTLY_IN_NORTHREND; end
	-- The current state we're in
	TitanWG_State.state = TitanWG_STATE_NOT_BEEN_TO_NORTHREND;
	-- The owner function.  This returns who the current owner of WG is, or nil
	-- if we have no current information
	TitanWG_State.updateOwner = function() 
		if ( TitanWG_owner == nil and TitanWG_State.isInfoAccurate() and TitanWG_isInNorthrend() ) then
			-- See if we have the WG buff
			
			-- Otherwise, use map landmarks
			local continentIndex = GetCurrentMapContinent();
			local zoneIndex = GetCurrentMapZone();
			-- set the map to Northrend
			SetMapZoom(TitanWG_NORTHREND_INDEX);
			-- Get the information we need from it
			local name, description, textureIndex, x, y;
			for i=1,GetNumMapLandmarks() do
				name,_,textureIndex,_,_ = GetMapLandmarkInfo(i);
				if ( name == L["TITAN_WG_WINTERGRASP"] ) then
					if ( textureIndex == 48 ) then 
						TitanWG_owner = L["TITAN_WG_HORDE"];
						break;
					elseif ( textureIndex == 46 ) then
						TitanWG_owner = L["TITAN_WG_ALLIANCE"];
						break;
					elseif ( textureIndex == 101 ) then
						break;
					else
						TitanWG_debug("textureIndex="..textureIndex);
					end
				end
			end
			-- set the map back
			SetMapZoom(continentIndex, zoneIndex, 1);
			if ( TitanWG_owner == nil ) then
				TitanWG_debug("Current owner: nil");
			else
				TitanWG_debug("Current owner: "..TitanWG_owner);
			end
		end
	end
	-- This function returns a string containing information about WG
	TitanWG_State.description =	function()
		if TitanWG_State.isInfoAccurate() then
			return L["TITAN_WG_CONTROLLER"]..":\t"..TitanWG_colourFactionText(TitanWG_owner).."\n"..
					L["TITAN_WG_NEXT_BATTLE"]..":\t"..TitanWG_getColouredText(TitanWG_formatTime(TitanWG_State.time()), "white").."\n\n";
		else
			if TitanWG_State.state == TitanWG_STATE_BATTLE_UNDERWAY then
				return TitanWG_colourFactionText(L["TITAN_WG_BATTLE_UNDERWAY"]).."\n"..
						L["TITAN_WG_TIME_ELAPSED"]..":\t"..TitanWG_getColouredText(TitanWG_formatTime(TitanWG_State.time()), "white").."\n\n";
			else
				local dt = time() - TimeOfLastTimes[TitanWG_realm];
				local tmax = TitanWG_MAX_BATTLE_LENGTH + TitanWG_MAX_TIME_BETWEEN_BATTLES;
				local eps = floor(dt/tmax)*300;
				
				if ( eps < 3600 ) then
					 return TitanWG_getColouredText(L["TITAN_WG_NO_ZONE_INFO"], "yellow").."\n\n"..
						L["TITAN_WG_NEXT_BATTLE"]..":\t"..TitanWG_getColouredText("~"..TitanWG_formatTime(TitanWG_State.time()-2*eps), "white").."\n"..
						L["TITAN_WG_TIME_SINCE_LEFT_NR"]..":\t"..TitanWG_getColouredText(TitanWG_formatTime(dt), "white").."\n"..
						L["TITAN_WG_PROBABLE_ERROR"]..":\t"..TitanWG_getColouredText("\194\177"..TitanWG_formatTime(eps), "white").."\n\n";
				else
					 return TitanWG_getColouredText(L["TITAN_WG_NO_ZONE_INFO"], "yellow").."\n\n";
				end
			end
		end 
	end

-- ************************************************************************** --
-- ******************************** Titan Functions ************************* --
-- ************************************************************************** --

function TitanPanelTitanWGButton_OnLoad(self)
	self.registry = {
		id = TITAN_WG_ID,
		menuText = L["TITAN_WG_MENU_TEXT"],
		buttonTextFunction = "TitanPanelTitanWGButton_GetButtonText",
		tooltipTitle = L["TITAN_WG_TOOLTIP_TITLE"],
		tooltipTextFunction = "TitanPanelTitanWGButton_GetTooltipText",
		savedVariables = {
			ShowLabelText = 1,
			EnableAnnouncing = false,
		}
	};	
	
	TitanWG_debug("Debugging output is turned on");
	self:RegisterEvent("PLAYER_ENTERING_WORLD");	
	self:RegisterEvent("VARIABLES_LOADED");		 
	TitanWG_updateState();
	TitanWG_debug("state: "..TitanWG_State.state);
	AceTimer.ScheduleRepeatingTimer("TitanPanel"..TITAN_WG_ID, TitanPanelPluginHandle_OnUpdate, 1, {TITAN_WG_ID, TITAN_PANEL_UPDATE_ALL })
	TitanWG_output(TITAN_WG_ID.." by Rothpada (Shadow Council)");
end 

function TitanPanelTitanWGButton_OnEvent(self, event)
	if ( event == "PLAYER_ENTERING_WORLD" ) then
		TitanWG_continent = GetCurrentMapContinent();
		TitanWG_player,_ = UnitName("player");
		TitanWG_realm = GetRealmName();
		TitanWG_debug("Current continent: "..TitanWG_continent);
	elseif ( event == "VARIABLES_LOADED" ) then
		TitanWG_debug("Loaded saved variables");
		
		if ( not TimeOfLastTimes ) then 
			TimeOfLastTimes = {}; 
			TitanWG_debug("Initialized TimeOfLastTimes"); 
		end
		if ( not TimeOfLastTimes[GetRealmName()] ) then 
			TimeOfLastTimes[GetRealmName()] = -1; 
			TitanWG_debug("Initialized TimeOfLastTimes["..GetRealmName().."]"); 
		end
		if ( not LastTimesSeen ) then 
			LastTimesSeen = {}; 
			TitanWG_debug("Initialized LastTimesSeen"); 
		end
		if ( not LastTimesSeen[GetRealmName()] ) then 
			LastTimesSeen[GetRealmName()] = -1; 
			TitanWG_debug("Initialized LastTimesSeen["..GetRealmName().."]"); 
		end
		if ( not VoA ) then 
			VoA = {}; 
			TitanWG_debug("Initialized VoA");
		end
		if ( UnitLevel("player") >= TitanWG_LEVEL_THRESHOLD and 
						  ( not VoA[GetRealmName()] ) ) then 
			VoA[GetRealmName()] = {}; 
			TitanWG_debug("Initialized VoA["..GetRealmName().."]"); 
		end
		if ( UnitLevel("player") >= TitanWG_LEVEL_THRESHOLD and 
						( not VoA[GetRealmName()][UnitName("player")] ) ) then 
			VoA[GetRealmName()][UnitName("player")] = {}; 
			TitanWG_debug("Initialized VoA["..GetRealmName().."]["..UnitName("player").."]");
		end
	end
end

function TitanPanelTitanWGButton_GetButtonText(id)
    local buttonText = "---";
	local warningMsg = "";
	
	TitanWG_updateState();
		
	local wgTime = TitanWG_State.time();
	
	if ( wgTime ~= nil ) then		
		if ( TitanWG_State.state == TitanWG_STATE_NOT_RECENTLY_IN_NORTHREND ) then
			-- We provide an upper bound in this case, the margin of error will
			-- be, on average, probably no more than 5 minutes per battle missed
			warningMsg = "<"..TitanWG_formatTime(wgTime);
			buttonText = TitanWG_getColouredText(warningMsg, "white");
		elseif ( TitanWG_State.state == TitanWG_STATE_NOT_BEEN_TO_NORTHREND ) then
			-- We provide an upper bound, though this can be extremely loose;
			-- the margin of error will likely be around 5m per battle missed
			local dt = time() - TimeOfLastTimes[TitanWG_realm];
			local tmax = TitanWG_MAX_BATTLE_LENGTH + TitanWG_MAX_TIME_BETWEEN_BATTLES;
			local eps = floor(dt/tmax)*300;
			
			if ( eps < 600 ) then
				warningMsg = "~"..TitanWG_formatTime(TitanWG_State.time());
				buttonText = TitanWG_getColouredText(warningMsg, "white");
			elseif ( eps < 1800 ) then
				warningMsg = "< "..TitanWG_formatTime(TitanWG_State.time());
				buttonText = TitanWG_getColouredText(warningMsg, "white");
			elseif ( eps < 3600 ) then
				warningMsg = "<< "..TitanWG_formatTime(TitanWG_State.time());
				buttonText = TitanWG_getColouredText(warningMsg, "white");
			end
		elseif ( TitanWG_State.state == TitanWG_STATE_RECENTLY_IN_NORTHREND ) then
			-- This is an estimate, though it should be quite accurate.  Also,
			-- our ownership information is still accurate
			warningMsg = "~"..TitanWG_formatTime(wgTime);
			buttonText = TitanWG_colourFactionText(warningMsg);
		elseif ( TitanWG_State.state == TitanWG_STATE_BATTLE_UNDERWAY ) then
			-- Give the time elapsed since the battle started
			warningMsg = TitanWG_formatTime(wgTime);
			buttonText = TitanWG_getColouredText(warningMsg, "yellow");
		elseif ( TitanWG_State.state == TitanWG_STATE_CURRENTLY_IN_NORTHREND) then
			-- This is the time reported by Blizzard's API function
			warningMsg = TitanWG_formatTime(wgTime);
			buttonText = TitanWG_colourFactionText(warningMsg);
		end
		-- Check the auto-announce, if enabled and in a guild and our time is
		-- going to be reasonably accurate
		if ( IsInGuild() and TitanWG_State.isInfoAccurate() ) then
			if ( TitanGetVar(TITAN_WG_ID, "EnableAnnouncing") and wgTime == TitanWG_WARNING_TIME and ( not TitanWG_announced ) ) then
				SendChatMessage(L["TITAN_WG_TIME_UNTIL_WG"]..": "..warningMsg,"GUILD");
				TitanWG_announced = true; -- Makes sure this only triggers once
			elseif ( wgTime < TitanWG_WARNING_TIME ) then
				TitanWG_announced = false;
			end
		end
	end

    return L["TITAN_WG_BUTTON_LABEL"], buttonText;
end

function TitanPanelTitanWGButton_GetTooltipText(self)
	local tooltip = "\n"..TitanWG_State.description();

    -- Show number of WG marks
    -- Show number of shards

    tooltip = tooltip..TitanWG_getColouredText(L["TITAN_WG_WINTERGRASP_CURRENCY_INFO"], "white")..":\n";
	local noCurrencies = true;
	for i=1,GetCurrencyListSize() do
		name, _, _, _, _, count, _, _, itemID = GetCurrencyListInfo(i);
		if ( itemID == 43228 or itemID == 43589 ) then
			noCurrencies = false;
			tooltip = tooltip..format("     %s:\t%s\n", TitanWG_getColouredText(name, "yellow"), TitanWG_getColouredText(count, "white"));
		end
	end
	
	if noCurrencies then
		tooltip = tooltip.."     "..TitanWG_getColouredText(L["TITAN_WG_NO_CURRENCIES"], "red").."\n";
	end

	if ( UnitLevel("player") >= TitanWG_LEVEL_THRESHOLD ) then
		-- Show if character is saved to Vault 10 or 25		
		VoA[TitanWG_realm][TitanWG_player]["normal"] = nil;
		VoA[TitanWG_realm][TitanWG_player]["heroic"] = nil;
	
		for i=1,GetNumSavedInstances() do
			local name, ID, expiry, difficulty = GetSavedInstanceInfo(i);
			expiry = time() + expiry;
			if ( name == L["TITAN_WG_VAULT_OF_ARCHAVON"] ) then
				if ( difficulty > 1 ) then
					VoA[TitanWG_realm][TitanWG_player]["heroic"] = expiry;
				else
					VoA[TitanWG_realm][TitanWG_player]["normal"] = expiry;
				end
			end
		end
	end
	
	tooltip = tooltip.."\n"..TitanWG_getColouredText(L["TITAN_WG_VAULT_OF_ARCHAVON_INFO"], "white")..":\n";
	
	for realm,players in pairs(VoA) do
		tooltip = tooltip.."   "..TitanWG_getColouredText(realm, "daa520").."\n";
		for player,raids in pairs(VoA[realm]) do
			if ( VoA[realm][player]["normal"] ~= nil and time() > VoA[realm][player]["normal"] ) then
				VoA[realm][player]["normal"] = nil;
			end
			if ( VoA[realm][player]["heroic"] ~= nil and time() > VoA[realm][player]["heroic"] ) then
				VoA[realm][player]["heroic"] = nil;
			end
			
			local v10t = "";
			local v25t = "";
			if ( VoA[realm][player]["normal"] == nil ) then
				v10t = TitanWG_getColouredText(L["TITAN_WG_UNSAVED"], "green");
			else
				v10t = TitanWG_getColouredText(L["TITAN_WG_SAVED"], "red");
			end
			if ( VoA[realm][player]["heroic"] == nil ) then
				v25t = TitanWG_getColouredText(L["TITAN_WG_UNSAVED"], "green");
			else
				v25t = TitanWG_getColouredText(L["TITAN_WG_SAVED"], "red");
			end
			-- TODO: Colour the current player differently
			if ( player == UnitName("player") ) then
				tooltip = tooltip..format("      %s - \t10: %s 25: %s\n", 
										TitanWG_getColouredText(player, "ffa500"),
										v10t, v25t
									 );
			else
				tooltip = tooltip..format("      %s - \t10: %s 25: %s\n", 
										TitanWG_getColouredText(player, "yellow"),
										v10t, v25t
									 );
			end	
		end
	end
	
    return tooltip;
end

function TitanPanelRightClickMenu_PrepareTitanWGMenu()
	local info;
	if ( UIDROPDOWNMENU_MENU_LEVEL == 2 ) then
		if ( UIDROPDOWNMENU_MENU_VALUE == "TITAN_WG_BROADCAST" ) then
			info = {};
			info.text = L["TITAN_WG_SAY"];
			info.func = function() TitanWG_broadcast("SAY"); end
			UIDropDownMenu_AddButton(info,2);
			info = {};
			info.text = L["TITAN_WG_YELL"];
			info.func = function() TitanWG_broadcast("YELL"); end
			UIDropDownMenu_AddButton(info,2);
			info = {};
			info.text = L["TITAN_WG_GUILD"];
			info.func = function() TitanWG_broadcast("GUILD"); end
			UIDropDownMenu_AddButton(info,2);
			info = {};
			info.text = L["TITAN_WG_PARTY"];
			info.func = function() TitanWG_broadcast("PARTY"); end
			UIDropDownMenu_AddButton(info,2);
			info = {};
			info.text = L["TITAN_WG_RAID"];
			info.func = function() TitanWG_broadcast("RAID"); end
			UIDropDownMenu_AddButton(info,2);
			local list = {GetChannelList()};
			for i=1,getn(list)/2 do
				info = {};
				info.text = list[i*2-1]..". "..list[i*2];
				info.func = function() TitanWG_broadcast("CHANNEL",list[i*2-1]); end
				UIDropDownMenu_AddButton(info,2);
			end
			
			TitanPanelRightClickMenu_AddSpacer(2);
			
			info = {};
			info.text = L["TITAN_WG_WHISPER"];
			info.func = function()
				StaticPopupDialogs["TITAN_WG_WHISPER_DIALOG"] = {
				  text = L["TITAN_WG_WHISPER_TEXT"],
				  button1 = L["TITAN_WG_ACCEPT"],
				  button2 = L["TITAN_WG_CANCEL"],
				  OnAccept = function() 
				  	TitanWG_broadcast("WHISPER", getglobal(this:GetParent():GetName().."EditBox"):GetText());
				  end,
				  timeout = 0,
				  whileDead = 1,
				  hideOnEscape = 1,
				  hasEditBox = 1
				};
				StaticPopup_Show("TITAN_WG_WHISPER_DIALOG");
			end
			UIDropDownMenu_AddButton(info,2);
		end
	else
		TitanPanelRightClickMenu_AddTitle(TitanPlugins[TITAN_WG_ID].menuText);
		TitanPanelRightClickMenu_AddTitle(L["TITAN_WG_CHANNEL"], 2); 
		TitanPanelRightClickMenu_AddTitle(L["TITAN_WG_AUTO_ANNOUNCE"], 2); 
		TitanPanelRightClickMenu_AddSpacer();
		
		info = {};
		info.text = L["TITAN_WG_BROADCAST"];
		info.value = "TITAN_WG_BROADCAST";
		info.hasArrow = 1;
		UIDropDownMenu_AddButton(info);
		if IsInGuild() then
			TitanPanelRightClickMenu_AddToggleVar(L["TITAN_WG_AUTO_ANNOUNCE"], TITAN_WG_ID, "EnableAnnouncing",2);
		end
				
		TitanPanelRightClickMenu_AddSpacer();
		TitanPanelRightClickMenu_AddToggleLabelText(TITAN_WG_ID);
		TitanPanelRightClickMenu_AddSpacer();
		TitanPanelRightClickMenu_AddCommand(L["TITAN_WG_HIDE"], TITAN_WG_ID, TITAN_PANEL_MENU_FUNC_HIDE);
	end
end

-- ************************************************************************** --
-- ******************************** User Functions ************************** --
-- ************************************************************************** --

function TitanWG_colourFactionText(msg)
	local colour = "white";
	if ( TitanWG_State.isInfoAccurate() ) then
		if ( TitanWG_State.time() < 0 ) then
			colour = "yellow";
		elseif ( TitanWG_owner ~= nil ) then
			if ( TitanWG_owner == UnitFactionGroup("player") ) then
				colour = "green";
			else  
				colour = "red";
			end
		end
	end
	
	return TitanWG_getColouredText(msg, colour);
end

function TitanWG_isInNorthrend()
	return TitanWG_continent == TitanWG_NORTHREND_INDEX and not IsInInstance();
end

function TitanWG_updateState()
	local oldState = TitanWG_State.state;
	-- Update the state that we're in
	if TitanWG_isInNorthrend() then
		if GetWintergraspWaitTime() then		
			TitanWG_State.state = TitanWG_STATE_CURRENTLY_IN_NORTHREND;
		else
			TitanWG_owner = nil;
			TitanWG_State.state = TitanWG_STATE_BATTLE_UNDERWAY;
		end
	elseif ( TitanWG_State.state == TitanWG_STATE_BATTLE_UNDERWAY ) then
		-- We consider our information stale because we don't know how long
		-- the battle will take, only an upper bound on that time
		TitanWG_owner = nil;
		TitanWG_State.state = TitanWG_STATE_NOT_RECENTLY_IN_NORTHREND;
	elseif ( TitanWG_State.state == TitanWG_STATE_CURRENTLY_IN_NORTHREND ) then
		-- We only left NR recently so our information is still accurate
		TitanWG_State.state = TitanWG_STATE_RECENTLY_IN_NORTHREND;
	elseif ( TitanWG_State.state == TitanWG_STATE_RECENTLY_IN_NORTHREND ) then
		-- We're not in NR.  See if we estimate that a battle has started and 
		-- must have ended
		if ( TitanWG_State.time() <= -TitanWG_MAX_BATTLE_LENGTH ) then
			-- If so, we say we've not recently been in NR and can provide at
			-- best an upper bound
			TitanWG_owner = nil;
			TitanWG_State.state = TitanWG_STATE_NOT_RECENTLY_IN_NORTHREND;
		end
	end
	if ( oldState ~= TitanWG_State.state ) then
		TitanWG_debug("old state: "..oldState..",new state: "..TitanWG_State.state);
	end
								
	TitanWG_State.updateOwner();	
    -- Find out how much time is left
	if ( TitanWG_State.state == TitanWG_STATE_NOT_RECENTLY_IN_NORTHREND or
			TitanWG_State.state == TitanWG_STATE_NOT_BEEN_TO_NORTHREND ) then
		-- Provide an upper bound
		TitanWG_State.time = function() return TitanWG_getUpperBound(); end
	elseif ( TitanWG_State.state == TitanWG_STATE_RECENTLY_IN_NORTHREND ) then -- We're not in NR so we estimate how much time we have left
			TitanWG_State.time = function() 
				return LastTimesSeen[TitanWG_realm] - (time() - TimeOfLastTimes[TitanWG_realm]); 
			end
	else -- We are in NR so GetWintergraspWaitTime will work
		if ( TitanWG_State.state == TitanWG_STATE_CURRENTLY_IN_NORTHREND ) then -- we are between battles
			 TitanWG_State.time = function()
				LastTimesSeen[TitanWG_realm] = GetWintergraspWaitTime();
				TimeOfLastTimes[TitanWG_realm] = time();
				return LastTimesSeen[TitanWG_realm];
			end
		else -- the battle is underway, so return how long the battle's been going for
			TitanWG_State.time = function() 
				return abs(LastTimesSeen[TitanWG_realm] - (time() - TimeOfLastTimes[TitanWG_realm])); 
			end
		end
	end
end

function TitanWG_debug(msg)
	if TitanWG_IS_DEBUGGING then
		TitanDebug("TitanWG> "..msg);
	end
end

function TitanWG_output(msg)
    DEFAULT_CHAT_FRAME:AddMessage(msg, 1, 1, 1);
end

function TitanWG_formatTime(timeInSeconds)
	return TitanUtils_GetAbbrTimeText(timeInSeconds);
end

function TitanWG_getColouredText(msg, colour)
	local colourCode = "";
	
	if ( msg == nil ) then
		return "";
	end
		
	if ( colour == nil or colour == "default" ) then
		return msg;
	elseif ( colour == "red" ) then
		colourCode = "|cffff0000";
	elseif ( colour == "green" ) then
		colourCode = "|cff00ff00";
	elseif ( colour == "blue" ) then
		colourCode = "|cff0000ff";
	elseif ( colour == "yellow" ) then
		colourCode = "|cffffff00";
	elseif ( colour == "white" ) then
		colourCode = "|cffffffff";
	elseif ( colour == "black" ) then
		colourCode = "|cff000000";
	else
		colourCode = "|cff"..colour;
	end
        
	return colourCode..msg..FONT_COLOR_CODE_CLOSE;
end

function TitanWG_getUpperBound()
	if ( TimeOfLastTimes[TitanWG_realm] == nil ) then
		return nil;
	end
	-- We provide an upper bound if we've not recently been to Northrend
	-- We also provide one if we've not ever been to NR, tho this is likely a WAG
	local bound;
	local tmax = TitanWG_MAX_BATTLE_LENGTH + TitanWG_MAX_TIME_BETWEEN_BATTLES;
	local tn = time();
	local t0 = TimeOfLastTimes[TitanWG_realm] + LastTimesSeen[TitanWG_realm];
	
	if ( t0 < tn ) then
		bound = tmax - mod(tn - t0, tmax); 	
	else
		bound = t0 - tn;
	end
	
	return bound;
end

function TitanWG_broadcast(channel,target)
	local wgTime = TitanWG_State.time();
	if ( wgTime ~= nil ) then
		SendChatMessage(L["TITAN_WG_TIME_UNTIL_WG"]..": "..TitanWG_formatTime(wgTime),channel,nil,target);
	end
end
